/* server.c */

#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include "shm.h"
#include "server.h"
#include "sock.h"

#ifdef SAAGAR
#define MSG_WAITALL 0
#endif

#define VERSION 1.1
#define MIN_X		21
#define MAX_X		59
#define MIN_Y		2
#define MAX_Y 		21

/* function prototypes */
void usage(void);
void doGame(int newsock);
void addPlayer();
void cleanUp();

/* global variables */
int sock = 0;
int newsock = 0;
int sem = 0;
struct Table *table = NULL;
int pid = -1;			/* player's id */
int verbose = 0;

int main(int argc, char *argv[]) {
  int i;
  int port = 0;
  int c;
  int shmkey, semkey;
  extern int optind, opterr;
  extern char *optarg;

  opterr = 1; /* set to 0 to disable error message */

  /* required: one of {-b or -p}, optional: -i<name> and/or -o<name> */
  while ((c = getopt (argc, argv, "hvp:")) != EOF) {
    switch (c) {
    case '?': /* returns ? if there is a getopt error */
    case 'h':
      usage();
      break;
    case 'v':
      verbose = 1;
      printf("Verbose mode on.\n");
      break;
    case 'p':
      port = atoi(optarg);
      break;
    }

  }
  if (port == 0) {
    printf("You must specify a port.\n");
    usage();
  }

  printf("Server activating...\n");

  /* setup shared memory */
  printf("Creating shm and sem...\n");
  shmkey = port;
  table = (struct Table *) shmcreate(shmkey, sizeof(struct Table));
  if (table == NULL) {
    perror("create");
    exit(1);
  }
  semkey = port + 1;
  if ((sem = semcreate(semkey, 1)) < 0) {
    perror("semcreate");
    exit(1);
  } 
  table->player_count = 0;
  table->paddlex = (MAX_X - MIN_X) / 2 + MAX_X;
  table->paddley = MAX_Y - 1;
  for (i=0; i < MAX_PLAYERS; i++)
    table->player[i].allocated = 0;
   
  /* create socket */
  printf("Creating socket...\n", argv[0]);
  sock = sockcreate(port);
  if (sock == -1) {
    perror("sockcreate");
    exit(1);
  }

  signal(SIGINT, cleanUp);	/* Ctrl-C to exit */

  while (1) {
    
    /* accept connections */
    newsock = sockaccept(sock);
    if (newsock < 0) {
      perror("sockaccept");
      exit(1);
    }

    printf("Connection request received.\n");
    if (fork() == 0) {

      /* child */
      printf("forked child.\n");
      close(sock);		/* don't need old socket */
      addPlayer();
      doGame(newsock);
      cleanUp();
      exit(0);

    } else {
      /* parent does nothing but loop back and accept more connections */
    }

  }
  
  exit(0);
}

/* if game full, cleanup else add player to game and return 0 */
void addPlayer() {
  extern int pid;
  extern struct Table *table;
  extern int sem;
  extern int newsock;
  struct textMessage buff;
  int i;

  /* Hmm.  I really need to check and add in one c.r.  Fix later. */
  semwait(sem);
  if (table->player_count == MAX_PLAYERS) {
    semsignal(sem);
    sprintf(buff.note, "%d", -1);
    puts("Game is full. Go away");
    if (send(newsock, &buff, sizeof(struct textMessage), 0) == -1) 
      perror("send");
    cleanUp();
  }
  semsignal(sem);

  /* get name */
  if (recv(newsock, &buff, sizeof(struct textMessage), MSG_WAITALL) < 0) {
    perror("recv");
    cleanUp();
  }
  if (buff.mtype != MESSAGE) {
    printf("Unexpected message type: %d\n", buff.mtype);
    cleanUp();
  }
  printf("Received name: '%s'\n", buff.note);

  /* add to table */
  semwait(sem);
  for (i=0; i < MAX_PLAYERS; i++)
    if (!table->player[i].allocated)
      break;
  pid = i;
  if (verbose)
    printf("Pid is %d\n", pid);
  strcpy(table->player[pid].name, buff.note);
  table->player[pid].score = 0;
  table->player[pid].allocated = 1;
  if (table->player_count == 0) { /* first player? */
    printf("First player\n");
    table->ballx = (MAX_X - MIN_X) / 2 + MIN_X;
    table->bally = (MAX_Y - MIN_Y) / 2 + MIN_Y + 2;
    table->paddlex = (MAX_X - MIN_X) / 2 + MIN_X;
    table->paddley = MAX_Y - 1;
    table->bally_vel = -1;
    table->turn = pid;
  }
  table->player_count++;
  semsignal(sem);
  sprintf(buff.note, "%d", pid);
  buff.mtype = MESSAGE;
  if (send(newsock, &buff, sizeof(struct textMessage), 0) == -1) {
    perror("send");
    cleanUp();
  }

  printf("Added to table.\n");

  return;
}

/* cleanUp */
void cleanUp() {
  extern struct Table *table;
  extern int sem;
  extern int sock;
  extern int newsock;
  extern int pid;
  int i;

  /* clean up */
  if (pid != -1) {
    printf("Player %d exiting.", pid);
    semwait(sem);
    /* if my turn, change it */
    if (table->turn == pid) {
      i = (table->turn + 1) % MAX_PLAYERS;
      while (1) {
	if (table->player[i].allocated)
	  break;
	i = (i + 1) % MAX_PLAYERS;
      }
      table->turn = i;
    }
    table -> player[pid].allocated = 0;
    table -> player_count--;
    semsignal(sem);
  }
  if (sock)
    close(sock);
  if (newsock)
    close(newsock);
  if (sem != -1) 
    semdelete(sem);
  if (table)
    shmdelete(table);
  printf("\nCleaned up.\n");
  exit(0);
}
  
/* play the main game */
void doGame(int newsock) {
  extern int verbose;
  struct Message buff;
  struct textMessage msg;
  int bytes;
  int i;

  bytes = recv(newsock, &buff, sizeof(struct Message), MSG_WAITALL);
  if (verbose)
    printf("Bytes: %d, Received message %d.\n", bytes, buff.mtype);
    
  while (bytes > 0) {
  
    switch(buff.mtype) {

    case MOVEBALL:
      if (verbose) {
	printf("received MOVEBALL\n");
	printf("Ok on ballmove.  x: %d, y:%d, xvel: %d, yvel: %d\n", 
	       buff.x, buff.y, buff.x_vel, buff.y_vel);
      }
      semwait(sem);
      table->ballx = buff.x;
      table->bally = buff.y;
      table->ballx_vel = buff.x_vel;
      table->bally_vel = buff.y_vel;
      semsignal(sem);
      msg.mtype = MESSAGE;
      strcpy(msg.note, "Ok");
      if (send(newsock, &msg, sizeof(struct textMessage), 0) == -1) {
	perror("send");
	cleanUp();
      }
      break;

    case MOVEPADDLE:
      if (verbose) {
	printf("received MOVEPADDLE\n");
	printf("Ok on paddlemove.  x: %d, y:%d \n", buff.x, buff.y);
      }
      semwait(sem);
      table->paddlex = buff.x;
      semsignal(sem);
      msg.mtype = MESSAGE;
      strcpy(msg.note, "Ok");
      if (send(newsock, &msg, sizeof(struct textMessage), 0) == -1) {
	perror("send");
	cleanUp();
      }
      break;

    case SENDGAME:
      if (verbose)
	printf("received SENDGAME\n");
      semwait(sem);
      if (send(newsock, table, sizeof(struct Table), 0) == -1) {
	perror("send");
	semsignal(sem);
	cleanUp();
      }
      semsignal(sem);
      break;

    case SWITCHTURN:
      if (verbose)
	printf("received SWITCHTURN\n");
      semwait(sem);
      i = (table->turn + 1) % MAX_PLAYERS;
      while (1) {
	if (table->player[i].allocated)
	  break;
	i = (i + 1) % MAX_PLAYERS;
      }
      table->turn = i;
      if (verbose) 
	printf("turn: %d\n", table->turn);
      semsignal(sem);
      msg.mtype = MESSAGE;
      strcpy(msg.note, "Ok");
      if (send(newsock, &msg, sizeof(struct textMessage), 0) == -1) {
	perror("send");
	cleanUp();
      }
      break;

    case SCORE:
      if (verbose)
	printf("received SCORE\n");
      semwait(sem);
      table->player[buff.x].score = table->player[buff.x].score + 1;
      table->ballx = (MAX_X - MIN_X) / 2 + MIN_X;
      table->bally = (MAX_Y - MIN_Y) / 2 + MIN_Y + 2;
      table->bally_vel = -1;
      table->ballx_vel = 0;
      i = (table->turn + 1) % MAX_PLAYERS;
      while (1) {
	if (table->player[i].allocated)
	  break;
	i = (i + 1) % MAX_PLAYERS;
      }
      table->turn = i;
      semsignal(sem);
      msg.mtype = MESSAGE;
      strcpy(msg.note, "Ok");
      if (send(newsock, &msg, sizeof(struct textMessage), 0) == -1) {
	perror("send");
	cleanUp();
      }
      break;

    case DOEXIT:
      if (verbose)
	printf("received DOEXIT\n");
      cleanUp();
      break;
    }

    bytes = recv(newsock, &buff, sizeof(struct Message), MSG_WAITALL);
    if (verbose)
      printf("Bytes: %d, Received message %d.\n", bytes, buff.mtype);

  } 

  printf("Connection closed without warning! Bytes: %d\n", bytes);
}

/* print a usage message and quit */
void usage(void) {
  fprintf(stderr, "Electronic Handball Server, v%.2f\n", VERSION);
  fprintf(stderr, "usage: server {-p #} [-v -h]\n");
  fprintf(stderr, "\t-p #\tport to listen on\n");
  fprintf(stderr, "\t-v\tverbose mode\n");
  fprintf(stderr, "\t-h\tthis usage message\n");
  exit(1);
}
